@@ -1,5 +1,6 @@
 import os
 from pathlib import Path
+from typing import Callable
 from typing import List
 from typing import Optional
 from typing import Sequence
@@ -7,8 +8,6 @@ from typing import Tuple
 from typing import Union
 
 from doorway._hash import HashAlgo
-from doorway._hash import HashMode
-from doorway._hash import hash_file
 from doorway._hash import hash_str
 
 
@@ -17,38 +16,41 @@ from doorway._hash import hash_str
 # ========================================================================= #
 
 
+_SHARD_KEYS = {
+    'basename': os.path.basename,
+    'abspath': os.path.abspath,
+    'input': lambda x: x,
+    None: lambda x: x,
+}
+
+ShardKey = Union[str, None, Callable[[str], str]]
+
 
 def shard_hash(
     file_path: Union[str, Path],
-    hash_data: str = 'name',
-    hash_mode: Optional[HashMode] = None,
+    shard_key: ShardKey = 'basename',
     hash_algo: Optional[HashAlgo] = None,
 ):
     assert isinstance(file_path, (str, Path)), f'file_path must be a str or Path, got type: {type(file_path)} with value: {repr(file_path)}'
     file_path = str(file_path)
+    # get the hash data function
+    if shard_key in _SHARD_KEYS:
+        shard_key = _SHARD_KEYS[shard_key]
+    elif not callable(shard_key):
+        raise ValueError(f'shard_key must be callable or one of: {list(_SHARD_KEYS.keys())}, got: {repr(shard_key)}')
     # compute the hash
-    if hash_data == 'name':
-        return hash_str(os.path.basename(file_path), hash_algo=hash_algo)
-    elif hash_data == 'abspath':
-        return hash_str(os.path.abspath(file_path), hash_algo=hash_algo)
-    elif hash_data == 'path':
-        return hash_str(file_path, hash_algo=hash_algo)
-    elif hash_data == 'disk':
-        return hash_file(file_path, hash_mode=hash_mode, hash_algo=hash_algo, hash_missing=False)
-    else:
-        raise KeyError(f'invalid hash_data: {repr(hash_data)}')
+    return hash_str(shard_key(file_path), hash_algo=hash_algo)
 
 
 def shard_idx(
     file_path: Union[str, Path],
     num_shards: int,
-    hash_data: str = 'name',
-    hash_mode: Optional[HashMode] = None,
+    shard_key: ShardKey = 'basename',
     hash_algo: Optional[HashAlgo] = None,
 ):
     assert isinstance(num_shards, int) and (num_shards > 0), f'num_shards must be an integer that is > 0, got: {repr(num_shards)}'
     # compute the hash for the file
-    hash = shard_hash(file_path, hash_data=hash_data, hash_mode=hash_mode, hash_algo=hash_algo)
+    hash = shard_hash(file_path, shard_key=shard_key, hash_algo=hash_algo)
     # convert hashes to integers, and assign to correct split
     idx = int(hash, 16) % num_shards
     return idx
@@ -69,8 +71,7 @@ _SHARD_RETURNS = {
 def sharded(
     file_paths: Sequence[Union[str, Path]],
     num_shards: int,
-    hash_data: str = 'name',
-    hash_mode: Optional[HashMode] = None,
+    shard_key: ShardKey = 'basename',
     hash_algo: Optional[HashAlgo] = None,
     returns: str = 'values'
 ) -> Union[List[List[int]], List[List[str]], List[List[Tuple[int, str]]]]:
@@ -85,7 +86,7 @@ def sharded(
     shards = [[] for _ in range(num_shards)]
     # assign paths to shards
     for i, file_path in enumerate(file_paths):
-        idx = shard_idx(file_path, num_shards, hash_data=hash_data, hash_mode=hash_mode, hash_algo=hash_algo)
+        idx = shard_idx(file_path, num_shards, shard_key=shard_key, hash_algo=hash_algo)
         shards[idx].append(value_getter(i, file_path))
     # results
     return shards
@@ -94,8 +95,7 @@ def sharded(
 def sharded_and_grouped(
     file_paths: Sequence[str],
     group_sizes: Sequence[int],
-    hash_data: str = 'name',
-    hash_mode: Optional[HashMode] = None,
+    shard_key: ShardKey = 'basename',
     hash_algo: Optional[HashAlgo] = None,
     returns: str = 'values'
 ) -> Union[List[List[int]], List[List[str]], List[List[Tuple[int, str]]]]:
@@ -111,8 +111,7 @@ def sharded_and_grouped(
     shards = sharded(
         file_paths=file_paths,
         num_shards=sum(group_sizes),
-        hash_data=hash_data,
-        hash_mode=hash_mode,
+        shard_key=shard_key,
         hash_algo=hash_algo,
         returns=returns,
     )
@@ -120,6 +119,7 @@ def sharded_and_grouped(
     splits, i = [], 0
     for size in group_sizes:
         splits.append([path for shard in shards[i:i+size] for path in shard])
+        i += size
     # done!
     return splits
 
